// aflpdata.C
//
using namespace std;

#include "util.h"
#include "graph.h"
#include "aflpdata.h"
#include <iostream>
#include <vector>
#include <string>

AFLPData::AFLPData(const string filename, const string markerfilename)
{
  ifstream f(filename.c_str());
  if(f.fail()) {
    cerr <<"Error: Cannot open data file " << filename << "." 
	 << endl;
    exit(1);
  }
  readAFLPFile(f);

  ifstream fm(markerfilename.c_str());
  if(fm.fail()) {
    cerr <<"Error: Cannot open data file " << markerfilename << "." 
	 << endl;
    exit(1);
  }
  readMarkerLength(fm);
}

int AFLPData::convChar(char c)
{
  if(c=='0')
    return 0;
  if(c=='1')
    return 1;
  if(c=='N' || c=='-')
    return MISSING;
  if(c=='\n')
    return END_OF_LINE;
  if(c==EOF)
    return END_OF_FILE;
  return ERROR;
}

int AFLPData::readMarker(ifstream& f,int& mark,char& bad)
{
  char c;
  do
    f.get(c);
  while (isspace(c) && !f.eof() && c!='\n');
  bad = c;
  return (mark = convChar(c));
}
    
int AFLPData::readName(ifstream& f,ostringstream& name)
{
  char c;
  f >> c;
  if(f.eof())
    return END_OF_FILE;
  name << c;

  // read in rest of name

  do {
    f.get(c);
    if(f.eof()) {
      cerr << "Read in end of file while reading first name." << endl;
      return ERROR;
    }
    if(c!=' ')
      name << c;
  } while (c!=' ');
  
  return 0;
} 

// *** int AFLP::readAFLPFile(ifstream& f) ***
// Assumes that the data for each taxon is on a single line.
// The first string with no spaces is the taxon name.
// Subsequent data is separated by non-end-of-line white space.
// If the number of sites for a subsequent taxon differs
//   from the first, it is an error.
// If a data character is not a `0', `1', `-', or `N', it is an error.

// Taxon names and markers data are read into temporary space first
// and then copied over to vectors/matrices of the correct size after the data
// is read in and the sizes are known.

int AFLPData::readAFLPFile(ifstream& f)
{
  numTaxa = 0;
  numMarkers = 0;

  ostringstream names[MAX_TAXA];
  int data[MAX_TAXA*MAX_MARKERS],index=0;
  char c;

  // read in first line of data
  
  int val = readName(f,names[numTaxa++]);

  if(val==ERROR || val==END_OF_FILE) {
    cerr << "Error: File ended while reading name of taxon " << numTaxa << "." << endl;
    exit(1);
  }
  int done=0;
  while (f && !done) {
    int mark;
    char bad;
    readMarker(f,mark,bad);
    switch(mark) {
    case ERROR :
      cerr << "Error: Taxon " << numTaxa << "(" << names[numTaxa-1].str() << ")" << ", Marker "
	   << numMarkers+1 << " is `" << bad << "', and should be `0,' `1', `-', or `N'." << endl;
      exit(1);
    case END_OF_FILE :
      cerr << "Error: File ends while reading first taxon." << endl;
      exit(1);
    case END_OF_LINE :
      done=1;
      break;
    default :
      if(numMarkers>= MAX_MARKERS) {
	cerr << "Error: Too many markers (" << numMarkers+1 << ") for taxon " << numTaxa << endl;
	exit(1);
      }
      data[index++] = mark;
      numMarkers++;
    }
  }
  
  // read in rest of data

  while (f) {
    int val = readName(f,names[numTaxa++]);
    if(val==ERROR) {
      cerr << "Error: File ended while reading name of taxon " << numTaxa << "." << endl;
      exit(1);
    }
    if(val==END_OF_FILE) {
      numTaxa--;
      break;
    }
    int mark;
    char bad;
    for(int j=0;j<numMarkers+1;j++) {
      readMarker(f,mark,bad);
      switch(mark) {
      case ERROR :
	cerr << "Error: Taxon " << numTaxa << "(" << names[numTaxa-1].str() << ")" << ", Marker "
	     << numMarkers+1 << " is `" << bad << "', and should be `0,' `1', `-', or `N'." << endl;
	exit(1);
      case END_OF_LINE :
	if(j < numMarkers) {
	  cerr << "Error: Taxon " << numTaxa << "(" << names[numTaxa-1].str() << ") has too few markers ("
	       << j << " instead of " << numMarkers << ")." << endl;
	  exit(1);
	}
	else
	  break;
      case END_OF_FILE :
	if(j==numMarkers)
	  break;
	else {
	  cerr << "Error: File ends prematurely.  Taxon " << numTaxa << "(" << names[numTaxa-1].str()
	       << ") has too few markers (" << j << " instead of " << numMarkers << ")." << endl;
	  exit(1);
	}
      default :
	if(j==numMarkers) {
	  cerr << "Error: Taxon " << numTaxa << "(" << names[numTaxa-1].str() << ") has too many markers." << endl;
	  exit(1);
	}
	else
	  data[index++] = mark;
      }
    }
  }
  taxaNames.resize(numTaxa);
  aflp.resize(numTaxa);
  for(int i=0;i<numTaxa;i++)
    aflp[i].resize(numMarkers);
  for(int i=0;i<numTaxa;i++) {
    taxaNames[i] = names[i].str();
    for(int j=0;j<numMarkers;j++)
      aflp[i][j] = data[numMarkers*i+j];
  }
  cout << "Read in " << numTaxa << " taxa with " << numMarkers << " markers each." << endl;
}

void AFLPData::readMarkerLength(ifstream& f)
{
  markerLength.resize(0);
  while (f){
    double tmp;
    f >> tmp;
    markerLength.push_back(tmp-39);
    char c;
    f.get(c);
    if(c=='\n') break;
  }
  if(markerLength.size() != numMarkers){
    cerr << "Error: Unequal mark size: markerLength.size = " << markerLength.size() << ", numMarkers = " << numMarkers << endl;
    exit(1);
  }
}


void AFLPData::makeDistMatrix(vector<vector<double> >& dist) const
{
  for(int i1=0; i1<numTaxa; i1++)
    dist[i1][i1]=0;

  int n11, n10;
  for(int i2=0; i2<numTaxa-1; i2++)
    for(int j=i2+1; j<numTaxa; j++){
      n11=0;
      n10=0;
      for(int k=0; k<numMarkers; k++)
	if(aflp[i2][k]==1 && aflp[j][k]==1)
	  n11++;
	else if(aflp[i2][k]==1 || aflp[j][k]==1)
	  n10++;
      dist[i2][j] = dist[j][i2] = double(n10) / double(n11+n10);
    }
}


void AFLPData::printDistMatrix(ostream& f, vector<vector<double> >& dist, int n) const
{
  string buf;
  f << setiosflags(ios::showpoint | ios::fixed);

  f << endl;
  f << setw(26) << " ";
  for(int i=0;i<n;i++)
    f << setw(6) << i+1 << " ";
  f << endl;
  f << setw(25) << " ";
  for(int i=0;i<n;i++)
    f << "-------";
  f << endl;
  
  for(int i=0;i<n;i++)
  {
    f << setw(2) << i+1 << " ";
    f << setw(22) << taxaNames[i] << " ";
    f << setprecision(4) << setw(6) << dist[i][0];
    for(int j=1;j<n;j++)
      f << " " << setprecision(4) << setw(6) << dist[i][j];
    f << endl;
  }
  f << endl;
}



void AFLPData::printSummary(ostream& f)
{
  vector<vector<int> > counts(numTaxa);
  for(int i=0;i<numTaxa;i++)
    counts[i].resize(3);

  for(int i=0;i<numTaxa;i++)
    for(int j=0;j<3;j++)
      counts[i][j] = 0;

  for(int i=0;i<numTaxa;i++)
    for(int j=0;j<numMarkers;j++) {
      int m = aflp[i][j];
      switch(m) {
      case -9 : counts[i][2]++; break;
      default : counts[i][m]++;
      }
    }

  f << endl;
  f << "*** Summary of AFLP marker counts ***" << endl << endl;
  f << numTaxa << " taxa with " << numMarkers << " markers each." << endl << endl;
  f << setw(20) << "Taxon" << setw(6) << 0 << setw(6) << 1 << setw(8) << "Missing" << endl;
  f << "-----------------------------------------" << endl;

  for(int i=0;i<numTaxa;i++) {
    f << setw(20) << taxaNames[i];
    for(int j=0;j<3;j++)
      f << setw(6) << counts[i][j];
    f << endl;
  }
  f << "-----------------------------------------" << endl << endl;

  f << endl << endl;
  f << "*** Distance Matrix ***" << endl << endl;
  vector<vector<double> > dist(numTaxa);
  for(int i=0; i<numTaxa; i++)
    dist[i].resize(numTaxa);
  makeDistMatrix(dist);
  printDistMatrix(f, dist, numTaxa);
}
