// graph.h
#ifndef GRAPHHDR
#define GRAPHHDR

#include "util.h"
#include "runsettings.h"
#include "aflpdata.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>

class AFLPData;
class Node;
class Edge {
public:
  Edge () {}
  Edge(int n) :
    number(n)
  {}
  int getNumber() { return number; }
  void setLength(double x) { length=x; }
  double getLength() { return length; }
  Node* getOtherNode(Node *n) { return (nodes[0] == n ? nodes[1] : nodes[0]); }
  void setNode(int i,Node *n) { nodes[i]=n; }
  void setNodes(Node *m,Node *n) { nodes[0]=m; nodes[1]=n; }
  Node* getNode(int n) { return nodes[n]; }
  void swapNodes() {
    Node *temp=nodes[0];
    nodes[0] = nodes[1];
    nodes[1] = temp;
  }
  void calculateTransitionMatricesM();
  void calculateTransitionMatricesC(double N);
  double getTProb2(int i,int j) { return tProb2[i][j]; }
  double getTProb16(int i,int j) { return tProb16[i][j]; }
  double getTProb18(int i,int j) { return tProb18[i][j]; }
  void setSavedLength();
  void setSavedProb2();
  void setBackLength();
  void setBackProb2();
  void print(ostream&);
private:
  int number;
  double length,savedLength;
  Node* nodes[2];
  double tProb2[2][2],savedTProb2[2][2];
  double tProb16[17][17],savedTProb16[17][17];
  double tProb18[19][19],savedTProb18[19][19];
};

class Node {
 public:
  Node() {}
  Node(int n,int l) : number(n), leaf(l) {}
  int getNumber() const { return number; }
  int isLeaf() { return leaf; }
  Edge* getEdge(int n) { return edges[n]; }
  void setEdge(int n,Edge *e) { edges[n] = e; }
  void swapEdges(int x,int y) {
    Edge *temp = edges[x];
    edges[x] = edges[y];
    edges[y] = temp;
  }
  Node* getNeighbor(int n) { return edges[n]->getOtherNode(this); }
  Node* getSib(){
    for(int i=1; i<3; i++){
      Node* parent=edges[0]->getNode(1);
      if(parent->getNeighbor(i) == this){
	return parent->getNeighbor(3-i);
	break;
      }
    }
  }
  double getSubLogProb16(int c,int m) { return subLogProb16[c][m]; }
  double getSubLogProb18(int c,int m) { return subLogProb18[c][m]; }
  void calculateSubtreeLogLikelihoods(const AFLPData&, const RunSettings&, int);
  void print(ostream&);
  void reOrderEdges(Node*);
  void reOrderEdgesRoot();
  void searchSibSubTree(int&, Node*, vector<int>&, int&, int);
  void searchSubTree(int&, Node*, vector<int>&, int&, vector<int>&);
  void travelSubTree(int&,vector<int>&);

 private:
  int number;
  int leaf;
  Edge *edges[3];
  double subLogProb16[2][17];
  double subLogProb18[2][19];
};

struct SavedStates{
  int NumSubEdges;
  Edge edges[6];
  Node nodes[7];
};


class Tree {
public:
  Tree(const RunSettings&,const AFLPData&,Rand&);
  Tree(void);
  ~Tree() {
    for(int i=0;i<numNodes;i++)
      delete nodes[i];
    for(int i=0;i<numEdges;i++)
      delete edges[i];
  }
  void setNodesAndEdges(int,double*);
  int getNumTaxa() { return numTaxa; }
  int getNumNodes() { return numNodes; }
  int getNumEdges() { return numEdges; }
  double getLogLikelihood() { return logLikelihood; }
  void setLogLikelihood(double x) { logLikelihood=x; }
  double calculateLogLikelihood(const AFLPData&, const RunSettings&, int);
  Node* getNode(int n) { return nodes[n]; }
  Edge* getEdge(int n) { return edges[n]; }
  void printNodesEdges() const;
  void findSmallest(vector<int>&,Node*) const;
  void printIntTop(ostream&,vector<int>&, Node*) const;
  void printIntBL(ostream&,vector<int>&, Node*) const;
  void printTop(ostream&) const;
  void printBL(ostream&) const;
  void print(ostream&) const;
  void updateTree(const RunSettings&,const AFLPData&, Rand&, int&, int&, double&);
  void createRandomTree(const RunSettings&,Rand&);
  void createGoodTree(const AFLPData&, Rand&);
  void reOrderEdges();
  void calculatePiMs();
  void calculatePiCs(int k, double N);
  void check(int ind[6][6]) const;

  void setTProb2(int k, int e, int i, int j, double x) { tProb2[k][e][i][j] = x; }
  void setTProb16(int e, int i, int j, double x) { tProb16[e][i][j] = x; }
  void setTProb18(int e, int i, int j, double x) { tProb18[e][i][j] = x; }
  void setSubLogProb16(int k, int n, int c, int i, double x) { subLogProb16[k][n][c][i] = x; }
  void setSubLogProb18(int k, int n, int c, int i, double x) { subLogProb18[k][n][c][i] = x; }
  void setPiM16(int i, double x) { piM16[i] = x; }
  void setPiM18(int i, double x) { piM18[i] = x; }
  void setPiC(int k, int i, double x) { piC[k][i] = x; }
  double getTProb2(int k, int e, int i, int j) { return tProb2[k][e][i][j]; }
  double getTProb16(int e, int i, int j) { return tProb16[e][i][j]; }
  double getTProb18(int e, int i, int j) { return tProb18[e][i][j]; }
  double getSubLogProb16(int k, int n, int c, int i) { return subLogProb16[k][n][c][i]; }
  double getSubLogProb18(int k, int n, int c, int i) { return subLogProb18[k][n][c][i]; }
  double getPiM16(int i) { return piM16[i]; }
  double getPiM18(int i) { return piM18[i]; }
  double getPiC(int k, int i) { return piC[k][i]; }
  void findNodesToRoot(Node*, vector<int>&, int&, int, int);
  void calculateSubLogProb(Node*, int, const AFLPData&, const RunSettings&);
  void setSavedProbs0();
  void setBackProbs0();
  //  double calculateNewLogLikelihood0(const AFLPData&, const RunSettings&);
  void setSavedProbs1(int edgeChosen);
  void setBackProbs1(int edgeChosen);
  double calculateNewLogLikelihood1(const AFLPData&, const RunSettings&);
  void setSavedProbs2();
  void setBackProbs2();
  double calculateNewLogLikelihood2(const AFLPData&, const RunSettings&);
  void setSavedProbs3();
  void setBackProbs3();
  double calculateNewLogLikelihood3(const AFLPData&, const RunSettings&);
  void setSavedValues4();
  void setBackValues4();
  double calculateNewLogLikelihood4(const AFLPData&, const RunSettings&);
  void setSavedValues5(int, int, int);
  void setBackValues5(int);
  double calculateNewLogLikelihood5(const AFLPData&, const RunSettings&, int);
  void setSavedValues6(int, int, vector<int>&, int);
  void setBackValues6();
  double calculateNewLogLikelihood6(const AFLPData&, const RunSettings&);


  // for check
  void setTProb2X(int k, int e, int i, int j, double x) { tProb2X[k][e][i][j] = x; }
  void setTProb16X(int e, int i, int j, double x) { tProb16X[e][i][j] = x; }
  void setTProb18X(int e, int i, int j, double x) { tProb18X[e][i][j] = x; }
  void setSubLogProb16X(int k, int n, int c, int i, double x) { subLogProb16X[k][n][c][i] = x; }
  void setSubLogProb18X(int k, int n, int c, int i, double x) { subLogProb18X[k][n][c][i] = x; }
  double getTProb2X(int k, int e, int i, int j) { return tProb2X[k][e][i][j]; }
  double getTProb16X(int e, int i, int j) { return tProb16X[e][i][j]; }
  double getTProb18X(int e, int i, int j) { return tProb18X[e][i][j]; }
  double getSubLogProb16X(int k, int n, int c, int i) { return subLogProb16X[k][n][c][i]; }
  double getSubLogProb18X(int k, int n, int c, int i) { return subLogProb18X[k][n][c][i]; }
  double calculateLogLikelihoodX(const AFLPData&, const RunSettings&);


private:
  int numTaxa;
  int numNodes;
  int numEdges;
  int numMarkers;
  vector<Node*> nodes;
  vector<Edge*> edges;
  vector<Node> savedNodes;
  vector<Edge> savedEdges;
  int root;
  SavedStates savedStates;
  double logLikelihood, savedLogLikelihood;
  double piM16[17];
  double piM18[19];
  vector<vector<double> > piC;                                              // numMarkers*2
  vector<vector<vector<vector<double> > > >  tProb2, savedTProb2;           // numMarkers*numEdges*2*2;
  vector<vector<vector<double> > > tProb16, savedTProb16;                   // numEdges*17*17
  vector<vector<vector<double> > > tProb18, savedTProb18;                   // numEdges*19*19
  vector<vector<vector<vector<double> > > > subLogProb16, savedSubLogProb16; // numMarkers*numEdges*2*17 
  vector<vector<vector<vector<double> > > > subLogProb18, savedSubLogProb18; // numMarkers*numEdges*2*19 
  vector<int> indNodesToRoot, indNodesToRoot2;
  vector<int> indEdgesToRoot;
  int numNodesToRoot;
  int numEdgesToRoot;
  vector<int> indNodesInPath, indNodesInPathNew;      // indNodesInPathNew is for the nodes ordered after update for calculate loglikelihood
  vector<int> indEdgesInPath;
  int numNodesInPath;
  int numEdgesInPath;
  int numNodesInPath2;                      // excluding root

  // for checking calculation
  vector<vector<vector<vector<double> > > >  tProb2X;           // numMarkers*numEdges*2*2;
  vector<vector<vector<double> > > tProb16X;                   // numEdges*17*17
  vector<vector<vector<double> > > tProb18X;                   // numEdges*19*19
  vector<vector<vector<vector<double> > > > subLogProb16X; // numMarkers*numEdges*2*17 
  vector<vector<vector<vector<double> > > > subLogProb18X; // numMarkers*numEdges*2*19 
  double logLikelihoodX;

  void connect2Nodes(Node*, Node*, double, double);
  Node *connect3Nodes(Node*,Node*,Node*, double, double);
  void searchPath(Node*, Node*);
};

#endif
