// main.C

// program to carry out a Bayesian analysis of AFLP marker data using MCMC
// current version assumes the Felsenstein restriction sites model which counts number of mismatches
// in each marker

// Bret Larget and Ruiyan Luo
// Sep 10, 2004

#include "util.h"
#include "runsettings.h"
#include "files.h"
#include "graph.h"
#include "aflpdata.h"
#include <iostream>
#include <vector>
#include <sys/times.h>

#define VERSION 1.01

using namespace std;

void printInitialInformation(Files& f, const RunSettings& rs, time_t& startTime)
{
  f.fout << "Output generated by aflp version " << VERSION << endl << endl;
  char name[257];
  gethostname(name,256);
  f.fout << "Started on " << name << " at " << ctime(&startTime);

  rs.print(f.fout);
  f.fout << endl;
  f.fout << "update methods: 0,1,2,3,4,5" << endl;
}

void updateIntervals(RunSettings& rs, Tree& t, Files& files, AcceptCounts& ac, int circ, int tuning)
{
  if(circ % rs.getSampleInterval() == 0) {
    files.flogp << t.getLogLikelihood() << endl;
    t.printTop(files.ftop);
    t.printBL(files.ftree);
    files.facc << setiosflags(ios::showpoint | ios::fixed) << setprecision(3);
    for(int i=0; i<7; i++) {
      files.facc << setw(5) << ac.moveNumSample[i] << setw(5) << ac.acceptNumSample[i] << setw(8)
		 << (ac.moveNumSample[i]>0 ? (double)ac.acceptNumSample[i]/(double)ac.moveNumSample[i] : 0.0);
      ac.moveNumSample[i] = ac.acceptNumSample[i] = 0;
    }
    files.facc << endl;
  }

  if(circ % rs.getWindowInterval() == 0) {
    cout << "Likelihood = " << t.getLogLikelihood() << endl;
    //    t.print(cout);
    t.printTop(cout);
    t.printBL(cout);
    cout << setiosflags(ios::showpoint | ios::fixed) << setprecision(3);
    for(int i=0; i<7; i++) {
      cout << setw(5) << ac.moveNumWindow[i] << setw(5) << ac.acceptNumWindow[i] << setw(8)
	   << (ac.moveNumWindow[i]>0 ? (double)ac.acceptNumWindow[i]/(double)ac.moveNumWindow[i] : 0.0);
      ac.moveNumWindow[i] = ac.acceptNumWindow[i] = 0;
    }
    cout << endl << endl;
    cout.flush();
  }

  if(tuning && (circ % rs.getTuningInterval() == 0)) {
      ac.acceptProbTuning[0] /= ac.moveNumTuning[0];
      if(ac.acceptProbTuning[0] < .2)
	rs.doubleAlphaA();
      else if(ac.acceptProbTuning[0] > .5)
	rs.halveAlphaA();
      ac.acceptProbTuning[1] /= ac.moveNumTuning[1];
      if(ac.acceptProbTuning[1] < .2)
	rs.doubleAlphaS();
      else if(ac.acceptProbTuning[1] > .5)
	rs.halveAlphaS();
      ac.acceptProbTuning[3] /= ac.moveNumTuning[3];
      if(ac.acceptProbTuning[3] < .2)
	rs.halveLambdaL();
      else if(ac.acceptProbTuning[3] > .5)
	rs.doubleLambdaL();
      ac.acceptProbTuning[4] /= ac.moveNumTuning[4];
      if(ac.acceptProbTuning[4] < .2)
	rs.halveLambdaG();
      else if(ac.acceptProbTuning[4] > .5)
	rs.doubleLambdaG();
      cout << setiosflags(ios::showpoint | ios::fixed) << setprecision(3);
      cout << "the averaged acceptance probabilities are: " << endl;
      cout << ac.acceptProbTuning[0] << setw(10) << ac.acceptProbTuning[1] << setw(10) << ac.acceptProbTuning[3] << setw(10) << ac.acceptProbTuning[4] << endl;
      cout << "the tuning parameters are: " << endl;
      cout << rs.getAlphaATuning() << setw(10) << rs.getAlphaSTuning() << setw(10) << rs.getLambdaLTuning() << setw(10) << rs.getLambdaGTuning() << endl;
      for(int i=0; i<7; i++){
	ac.moveNumTuning[i]=0;
	ac.acceptProbTuning[i]=0;
      }
  }	
}


void update(RunSettings& rs, AFLPData ad, Tree& t, Files& files, Rand& rand, int burn)
{
  int cycles = (burn ? rs.getPreBurnCycles() : rs.getNumCycles());
  int move, accept;
  double acceptProb;
  AcceptCounts ac;
  for(int i=0; i<7; i++){
    ac.moveNumSample[i]=0;
    ac.moveNumWindow[i]=0;
    ac.moveNumTuning[i]=0;
    ac.acceptNumSample[i]=0;
    ac.acceptNumWindow[i]=0;
    ac.acceptProbTuning[i]=0;
  }

  for(int circ=1; circ<=cycles; circ++){
    t.updateTree(rs, ad, rand, move, accept, acceptProb);
    ac.moveNumSample[move]++;
    ac.moveNumWindow[move]++;
    ac.moveNumTuning[move]++;
    ac.acceptNumSample[move]+=accept;
    ac.acceptNumWindow[move]+=accept;
    ac.acceptProbTuning[move]+=acceptProb;
    updateIntervals(rs, t, files, ac, circ, burn);
  }
}

void printEnd(Files& f, time_t& startTime, time_t& endTime)
{
    char name[257];
    gethostname(name,256);
    f.fout << "Finished on " << name << " at " << ctime(&endTime);
    int diff=endTime-startTime,days=diff/(24*60*60),hours=diff%(24*60*60)/(60*60);
    int minutes=diff%(60*60)/60,seconds=diff%60;
    f.fout << "Elapsed time: ";
    if(days>0)
      f.fout << days << (days==1 ? " day, " : " days, ");
    if(days>0 || hours>0)
      f.fout << hours << (hours==1 ? " hour, " : " hours, ");
    if(days>0 || hours>0 || minutes>0)
      f.fout << minutes << (minutes==1 ? " minute, " : " minutes, ");
    f.fout << seconds << (seconds==1 ? " second." : " seconds.") << endl;
}

int main(void)
{
  RunSettings rs(cin,"standard input"); // Read in the runsettings.
  Files files(rs);                      // Open input and output files.
  time_t startTime;
  time(&startTime);                     // Record time program begins.
  printInitialInformation(files,rs,startTime);    // Print initial information to the out file.
  int numChains = 1;
  double temperatureParam = 1;
  int numMethods = 6;
  Rand rand(rs.getSeed());              // Initialize the random number generator.
  AFLPData ad(rs.getDataFile(), rs.getMarkerFile());        // Read in AFLP marker data and marker lengths
  ad.printSummary(files.fout);          // Print summary of AFLP marker data
  Tree t(rs,ad,rand);                   // Create initial tree
  t.print(cout);
  t.printTop(files.ftop);
  t.printBL(files.ftree);
  t.calculateLogLikelihood(ad,rs,1);
  cout << "logLikelihood = " << t.getLogLikelihood() << endl;

  cout << endl;
  cout << "Starting burn in..." << endl;
  update(rs, ad, t, files, rand, 1);

  cout << "Entering main loop..." << endl;
  update(rs, ad, t, files, rand, 0);

  time_t endTime;
  time(&endTime);                     // Record time program begins.
  printEnd(files,startTime,endTime);

  return 0;
}
