//runsettings.C

#include "runsettings.h"

vector<RunSetting*> RunSetting::inputSettings;

void RunSetting::parseAll(const RunSettings& rs)
{
  for(vector<RunSetting*>::iterator i=inputSettings.begin();i!=inputSettings.end();i++)
    (*i)->parse(rs);
}

void RunSetting::printAll(ostream &c)
{
  for(vector<RunSetting*>::iterator i=inputSettings.begin();i!=inputSettings.end();i++) {
    if((*i)->shouldPrint())
      (*i)->print(c);
  }
}

void RunSetting::printDoc(ostream &c)
{
  c << endl;
  c << "Run settings are read from standard input, which may be redirected from a file." << endl;
  c << "Run settings are of the form <key>=<value>, one run setting per line." << endl;
  c << "The possible keys are:" << endl;
  c << "-------------------------------------------------------------------------------" << endl;
  c << setw(30) << left << "include-file" << " " << "file from which to read additional run settings" << endl;
  for(vector<RunSetting*>::iterator i=inputSettings.begin();i!=inputSettings.end();i++)
    c << setw(30) << left << (*i)->name << " " << (*i)->doc << endl;
  c << "-------------------------------------------------------------------------------" << endl;
}

void RunSetting::printDefaults(ostream &c)
{
  c << endl;
  c << "The default values for the run settings are:" << endl;
  c << "-------------------------------------------------------------------------------" << endl;
  c << "include-file" << " has no default setting" << endl;
  for(vector<RunSetting*>::iterator i=inputSettings.begin();i!=inputSettings.end();i++)
    c << (*i)->name << "=" << (*i)->val << endl;
  c << "-------------------------------------------------------------------------------" << endl;
}

void RunSetting::setValue(const string v, int i, const string fn) 
{
  if(line)
    error << "Warning: parameter (" << name << ") multiplied defined (line "
	  << line << " of " << fn << " and line " << i << " of " << fn
	  << "). Last value is used." << endError;
  line = i;
  filename = fn;
  val = v;
}

RunSetting* RunSetting::findInputSetting(const string s) 
{
  /* Returns a pointer to a parameter given the name s of a valid parameter.
     Otherwise returns NULL; */

  if(s=="")
    return NULL;
  vector<RunSetting*>::iterator i=inputSettings.begin();
  while(i!=inputSettings.end()) {
    if(s==(*i)->name)
      return *i;
    i++;
  }
  return NULL;
}

void RunSetting::parseError() 
{
  error << "Error: Could not understand parameter value '"
	<< val.substr(0,val.find_last_not_of(" \t\n")+1) << "' (line "
	<< line << " of " << filename << ")." << endError;
  quit(1);
}

void RunSetting::boundsError() 
{
  error << "Error: " << name << "(" << val << ") must be in the range " << bounds
	<< " (line " << line << " of " << filename << ")" << endError;
  quit(1);
}

void RunSetting::boundsParseError() 
{
  error << "Internal Error: Could not understand bounds " << bounds << " in run setting "
	<< name << "." << endl;
  quit(1);
}

template<class T> void RunSetting::checkBounds(T v) 
{
  if(bounds=="")
    return;

  int strict,sign=1;
  T n;
  char ch;
  char word[bounds.size()+1];
  ISTRSTREAM bd(bounds);

  if(!(bd >> ch))
    boundsParseError();
  if(ch=='[')
    strict = 0;
  else if(ch=='(')
    strict = 1;
  else 
    boundsParseError();

  if(!(bd >> ch))
    boundsParseError();
  if(ch=='-')
    sign = -1;
  else if(ch=='+')
    sign = 1;
  else
    bd.putback(ch);

  if(isdigit(ch) && (bd >> n)) {
    n *= sign;
    if(strict ? v <= n : v < n)
      boundsError();
  }
  else if(bd.read(word,3) && bd.gcount()==3 && toupper(word[0])=='I' 
	  && toupper(word[1])=='N' && toupper(word[2])=='F') {
      if(sign!=-1 || !strict)
	boundsParseError();
  }
  else 
    boundsParseError();

  if(!(bd >> ch) || (ch!=','))
    boundsParseError();

  if(!(bd >> ch))
    boundsParseError();
  if(ch=='-')
    sign = -1;
  else if(ch=='+')
    sign = 1;
  else
    bd.putback(ch);

  sign = 1;
  if(isdigit(ch) && (bd >> n)) {
    n *= sign;
    if(!(bd >> ch))
      boundsParseError();
    if(ch==']')
      strict = 0;
    else if(ch==')')
      strict = 1;
    else 
      boundsParseError();
    if(strict ? v >= n : v > n)
      boundsError();
  }
  else if(bd.read(word,3) && bd.gcount()==3 && toupper(word[0])=='I' 
	  && toupper(word[1])=='N' && toupper(word[2])=='F') {
    if(!(bd >> ch))
      boundsParseError();
    if(ch==']')
      strict = 0;
    else if(ch==')')
      strict = 1;
    else 
      boundsParseError();
    if(sign!=1 || !strict)
      boundsParseError();
  }
  else 
    boundsParseError();
}

class ParseTree {
private:
  enum Comp { EQ, NEQ, LEQ, REQ };

  class PTNode {
  public:
    vector<RunSetting*> rs;	// For a leaf, the list of runsettings.
    Comp comp;			// For a leaf, the comparison operation.
    vector<string> val;		// For a leaf, the list of values.
    char op;			// For an internal node or a compressd leaf, the operation.
    PTNode *left, *right;	// The children of the node.

    PTNode() : left(NULL), right(NULL) {}
    PTNode(PTNode *l, char c, PTNode *r) : op(c), left(l), right(r) {}
    PTNode(RunSetting *r, Comp c, string v) : comp(c), left(NULL), right(NULL) {
      rs.push_back(r);
      val.push_back(v);
    }
  };

public:
  ParseTree(string str) { string::iterator p=str.begin(); top = parseTreeInt(p,0,str); }
  int eval() { return evalInt(top); }
  void print(ostream& c) const { printInt(c,top,0,0,0); }
  void reduce() { top = reduceInt(top); } 

private:
  PTNode *top;
  PTNode *parseTreeInt(string::iterator&,int,const string&);
  int equal(const RunSetting *,const string);
  PTNode *addNode(PTNode *, char, PTNode *);
  int attachNode(PTNode *,char,PTNode *);
  int printInt(ostream&,const PTNode *,int,int,int) const;
  int count(PTNode *,char) const;
  int evalInt(PTNode *);
  PTNode *reduceInt(PTNode *);
};

ParseTree::PTNode *ParseTree::addNode(PTNode *p, char c, PTNode *q) 
{ 
  if(p->left==NULL && p->comp==EQ && attachNode(p,c,q))
    return q;
  else
    return new PTNode(p,c,q);
}

int ParseTree::attachNode(PTNode *p, char c, PTNode *q) 
{
  if(q->left!=NULL)
    if(c==q->op)
      return attachNode(p,c,q->left);
    else
      return 0;
  if((q->comp==EQ || (q->comp==REQ && q->op==c)) && p->rs[0]==q->rs[0] && c=='|') {
    q->op = c;
    q->comp = REQ;
    q->val.push_back(p->val[0]);
    return 1;
  }
  else if((q->comp==EQ || (q->comp==LEQ && q->op==c)) && (p->val[0] == q->val[0])) {
    q->op = c;
    q->comp = LEQ;
    q->rs.push_back(p->rs[0]);
    return 1;
  }
  else
    return 0;
}

int ParseTree::printInt(ostream &c, const PTNode *p, int prec, int number, int outOf) const
{
  if(p==NULL)
    return number;
  if(p->left==NULL) {// leaf
    switch(p->comp) {
    case EQ:
      c << p->rs[0]->getName() << " is " << p->val[0];
      break;
    case NEQ: 
      c << p->rs[0]->getName() << " is not " << p->val[0];
      break;
    case LEQ: {
      if(p->rs.size()==2)
	c << (p->op=='&' ? "both " : "either ") << p->rs[1]->getName()
	  << (p->op=='&' ? " and " : " or ") << p->rs[0]->getName();
      else { // p->rs.size()>2
	c << (p->op=='&' ? "all of " : "at least one of") << p->rs[p->rs.size()-1]->getName();
	for(int i=p->rs.size()-2;i>0;i--)
	  c << ", " << p->rs[i]->getName();
	c << ", and " << p->rs[0]->getName();
      }
      c << (p->op=='&' ? " are " : " is ") << p->val[0];
      break;
    }
    case REQ: 
      c << p->rs[0]->getName() << " is ";
      if(p->val.size()==2)
	c << "either " << p->val[1] << " or " << p->val[0];
      else {
	c << "one of " << p->val[p->val.size()-1];
	for(int i=p->val.size()-2;i>0;i--)
	  c << ", " << p->val[i];
	c << ", or " << p->val[0];
      }
      break;
    }
    return number+1;
  }
  else {
    int newPrec=(p->op=='&' ? 2 : 1);
    if(newPrec<prec) {
      c << "(";
      printInt(c,p,0,0,0);
      c << ")";
      return number+1;
    }
    else {
      int oldNumber=number;
      if(newPrec>prec) {
	number = 1;
	outOf = count(p->left,p->op) + count(p->right,p->op);
      }
      number = printInt(c,p->left,newPrec,number,outOf);
      if(p->op=='&')
	c << (outOf==2 ? " and " : (number==outOf ? ", and " : ", "));
      else
	c << (outOf==2 ? " or " : (number==outOf ? ", or " : ", "));
      number = printInt(c,p->right,newPrec,number,outOf);
      return (newPrec>prec ? oldNumber+1 : number);
    }
  }
}

int ParseTree::count(PTNode *p, char c) const
{
  return ((p->left==NULL || p->op!=c) ? 1 : count(p->left,c)+count(p->right,c));
}  

ParseTree::PTNode *ParseTree::reduceInt(PTNode *p)
{
  if(p->left==NULL) // leaf
    if(evalInt(p))
      return NULL;
    else if(p->comp!=LEQ || p->op!='&')
      return p;
    else {
      PTNode *q = new PTNode();
      q->op = '&';
      q->val.push_back(p->val[0]);
      for(int i=0;i<p->rs.size();i++)
	if(!equal(p->rs[i],p->val[0]))
	  q->rs.push_back(p->rs[i]);
      q->comp = (q->rs.size()==1 ? EQ : LEQ);
      return q;
    }
  else { // Internal node
    PTNode* l=reduceInt(p->left);
    PTNode* r=reduceInt(p->right);
    if(p->op=='&')
      return (l==NULL ? r : (r==NULL ? l : new PTNode(l,'&',r)));
    else
      return ((l==NULL || r==NULL) ? NULL : new PTNode(l,'|',r));
  }
}

int ParseTree::evalInt(PTNode *p) 
{
  if(p->left==NULL) // leaf
    switch(p->comp) {
    case EQ: return equal(p->rs[0],p->val[0]); 
    case NEQ: return !equal(p->rs[0],p->val[0]);
    case LEQ: {
      int b = equal(p->rs[0],p->val[0]);
      for(int i=1;i<p->rs.size();i++) {
	int c=equal(p->rs[i],p->val[0]);
	b = (p->op=='&' ? (b && c) : (b || c));
      }
      return b;
    }
    case REQ: {
      int b = equal(p->rs[0],p->val[0]);
      for(int i=1;i<p->val.size();i++)
	b = b || equal(p->rs[0],p->val[i]);
      return b;
    }
    }
  else if(p->op=='&')
    return evalInt(p->left) && evalInt(p->right);
  else
    return evalInt(p->left) || evalInt(p->right);
  return 0; // The compiler (-pedantic) wants this lines.
}

int ParseTree::equal(const RunSetting *rs, const string val)
{
  return (upcaseString(rs->getValue()) ==  upcaseString(val));
}

ParseTree::PTNode *ParseTree::parseTreeInt(string::iterator &p, int prec, const string &str)
{
  PTNode *b;
  while(isspace(*p))
    p++;      
  if(*p=='(') {
    p++;
    b = parseTreeInt(p,0,str);
    while(p!=str.end() && isspace(*p))
      p++;      
    if(*p!=')') {
      error << "Internal Error: Expected '(', found " << *p << " in " << str << endError;
      quit(1);
    }
    p++;
  }
  else {
    string::iterator p0=p;
    while(p!=str.end() && !isspace(*p) && *p!='!' && *p!='=')
      p++;
    RunSetting* rs=RunSetting::findInputSetting(string(p0,p));
    if(rs==NULL) {
      error << "Internal Error: Could not find run setting " << string(p0,p) << " in " << str << endError;
      quit(1);
    }
    while(p!=str.end() && isspace(*p))
      p++;      
    if(p+1<str.end() && (*p!='!' && *p!='=') || *(p+1)!='=') {
      error << "Internal Error: Expected '==' or '!=', found " << *p << ((*p=='!' || *p=='=') ? *(p+1) : ' ')
	    << " in " << str << endError;
      quit(1);
    }
    Comp comp=(*p=='!' ? NEQ : EQ);
    p += 2;
    while(p!=str.end() && isspace(*p))
      p++;      
    p0 = p;
    while(p!=str.end() && !isspace(*p) && *p!=')')
      p++;
    b = new PTNode(rs,comp,string(p0,p));
  }
  for(;;) {
    while(isspace(*p))
      p++;      
    if(p==str.end() || *p==')')
      return b;
    if(*p=='&' && *(p+1)=='&') {
      p += 2;
      b = addNode(b,'&',parseTreeInt(p,2,str));
    }
    else if(*p=='|' && *(p+1)=='|') {
      if(prec>1)
	return b;
      p += 2;
      b = addNode(b,'|',parseTreeInt(p,1,str));
    }
    else {
      error << "Internal Error: Expected ')', '&&', '||', or EOS; found " << *p << " in " << str << endError;
      quit(1);
    }
  }
}

int RunSetting::shouldPrint() const
{
  if(line>0 || restriction=="")
    return 1;
  ParseTree pt(restriction);
  return pt.eval();
}

void RunSetting::checkRestriction()
{
  if(line==0 || restriction=="")
    return;
  ParseTree pt(restriction);
  if(!pt.eval()) {
    pt.reduce();
    error << "Warning: Parameter " << name << " is only used when ";
    pt.print(error);
    error << " (line " << line << " of " << filename << ")." << endError;
  }
}

const string RSBoolean::names[] = {"false", "true"};

RSBoolean::RSBoolean(string n, string d, string v, string r="") : RSEnum(names,sizeof(names)/sizeof(string),n,d,v,r) {}

const string RSTreeType::names[] = {"random", "good", "file"};

RSTreeType::RSTreeType(string n, string d, string v, string r="") : RSEnum(names,sizeof(names)/sizeof(string),n,d,v,r) {}

const string RSUpdateMethod::names[] = {"1","1x","2","2x","3","4","5","5x"};

RSUpdateMethod::RSUpdateMethod(string n, string d, string v, string r="") : RSEnumList(names,sizeof(names)/sizeof(string),n,d,v,r) {}

void RunSetting::addGroup(int n, vector<int> &groups)
{
  vector<int>::iterator i=groups.begin();
  while(i!=groups.end() && *i<=n)
    i++;
  groups.insert(i,n);
}

void RunSetting::printBadParam(const string param, int lineno, const string fn)
{
  error << "'" << param.substr(0,param.find_last_not_of(" \t\n")+1)
	<<"' (line " << lineno << " of " << fn << ") ." << endError;
}

void RSGroupList::parse(const RunSettings &rs) 
{
  v = val.substr(0,val.find_last_not_of(" \t\n")+1);
  string::iterator sv = v.begin();
  if(!getGroups(sv,groups,val,line,filename)) {
    error << "Error: Empty group list ";
    printBadParam(val,line,filename);
    quit(1);
  }
  if(*sv == '*') {
    (sv)++;
    while(isspace(*sv))
      (sv)++;
  }
  if(sv != v.end()) {
    error << "Error: Group list ends badly ";
    printBadParam(val,line,filename);
    quit(1);
  }
  checkRestriction();
}

int RunSetting::getGroups(string::iterator &str, vector<int> &groups, string &param, 
			  int lineno, string &fn) 
{
  /* Parses a string representing a list of categories up to, but
     not including the final asterisk (if it exists).
     Checks for syntax errors and sets up the group array.
     str is the unscanned part of the group list string.
     groups is an array of categories (in order) seen so far.
     Returns whether or not a group was read in from the string. */

  int n;

  while(isspace(*str))
    str++;
  if(*str == '(') {
    str++;
    if(!getGroups(str,groups,param,lineno,fn)) {
      error << "Error: Empty group in group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    while(isspace(*str))
      str++;
    if(*str == '*') {
      error << "Error: Asterisk must be the last symbol in the group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    if(*str != ')') {
      error << "Error: Mismatched parentheses in group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    str++;
  }
  else if(isdigit(*str)) {
    n = *str - '0';
    for(str++;isdigit(*str);str++)
      n = 10 * n + *str - '0';
    if(n == 0) {
      error << "Error: Group must be a positive number in group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    else addGroup(n,groups);
  }
  else 
    return 0;
  while(isspace(*str))
    str++;
  if(*str == '^') {
    str++;
    while(isspace(*str))
      str++;
    if(!isdigit(*str)) {
      error << "Error: Repeat count must be an integer in the group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    n = *str - '0';
    for(str++;isdigit(*str);str++)
      n = 10 * n + *str - '0';
    if(n == 0) {
      error << "Error: Repeat count may not be zero in the group list ";
      printBadParam(param,lineno,fn);
      quit(1);
    }
    while(isspace(*str))
      str++;
  }
  if(*str == ',') {
    str++;
    getGroups(str,groups,param,lineno,fn);
  }
  return 1;
}

RunSettings::RunSettings(istream& f, const string& filename) :

  /* Reads in and processes the run settings from file f. Sets
     the corresponding object variables. A run setting not listed in the
     file is set to its default value. The run settings are processed
     in the order given in this constructor. */

  seed("seed","seed for random number generator","194024933"),
  preBurnCycles("pre-burn","number of pre-burn cycles","0","[0,Inf)"),
  cycles("cycles","number of cycles","10000","[0,Inf)"),
  sampleInterval("sample-interval","interval for storing the tree","100","(0,Inf)"),
  windowInterval("window-interval","interval for printing to the terminal","1000","(0,Inf)"),
  tuningInterval("tuning-interval","interval for changing the tuning parameter","1000","(0,Inf)"),
  dataFile("data-file","name of file in which data is stored","infile"),
  markerFile("marker-file","name of file in which marker lengths are stored","inMarkerfile"),
  outgroup("outgroup","outgroup used to root the tree for printing","1","(0,Inf)"),
  printAllTrees("print-all-trees","whether or not to print all the trees","true"),
  fileRoot("file-root","prefix for names of output files","run1"),
  initialTreeType("initial-tree-type","source of the input tree","random"),
  treeFile("tree-file","file containing initial tree","treefile","initial-tree-type==file"),
  mu("mu","expected edge length","0.5","(0,Inf)"),
  alphaATuning("alphaA-tune", "tuning parameter for updating all edge lengths", "10", "(0,Inf)"),
  alphaSTuning("alphaS-tune", "tuning parameter for updating single edge length", ".625", "(0,Inf)"),
  lambdaLTuning("lambdaL-tune", "tuning parameter for stretching factor in local deformation algorithm", ".05", "(0,Inf)"),
  lambdaGTuning("lambdaG-tune", "tuning parameter for stretching factor in generalized local deformation algorithm", ".05", "(0,Inf)")
  //  debug("debug","whether or not to print debugging statements","false"),
  //  debugCycle("debug-cycle","first cycle to being printing debugging stmts.","0","[0,Inf)","debug==true"),
  //  updateList("use-updates","list of update methods to use","1,1x,2,2x,3,4,5,5x"),
  //  mcmcmc("use-mcmcmc","whether or not to use MCMCMC","false"),
  //  numChains("chains","number of chains used in MCMCMC","4","[2,20]","use-mcmcmc==true"),
  //  swapFrequency("swap-frequency","frequency for swapping chains in MCMCMC","1","[1,Inf)","use-mcmcmc==true"),
  //  temperature("temperature","temperature for MCMCMC","0.005","(0,Inf)","use-mcmcmc==true")
{
  readFile(f,filename.c_str());
  RunSetting::parseAll(*this);
}

void RunSettings::readFile(istream& f, const string filename)
{
  string buf;
  int lineNum = 0;

  while(getline(f,buf)) {
    lineNum++;
    buf = buf.substr(0,buf.find_first_of("#"));
    buf = buf.substr(0,buf.find_last_not_of(" \t\n")+1);
    if(buf.empty())
      continue;
    buf = buf.substr(buf.find_first_not_of(" \t\n"));
    int eqpos=buf.find_first_of('=');
    if(eqpos!=string::npos) {
      string word1=buf.substr(0,eqpos),word2=buf.substr(eqpos+1);
      if(word1.empty()) {
	error << "Error: Missing run setting (line " << lineNum
	      << ") of " << filename << "." << endError;
	quit(1);
      }
      if(word2.empty()) {
	error << "Error: Missing run setting value (line " << lineNum
	      << ") of " << filename << "." << endError;
	quit(1);
      }
      word1 = word1.substr(0,word1.find_last_not_of(" \t\n")+1);
      word2 = word2.substr(word2.find_first_not_of(" \t\n"));
      if(word1=="include-file") {
	ifstream f(word2.c_str());
	if(f.fail()) {
	  error <<"Error: Cannot open run settings include file " << word2 << "." << endError;
	  quit(1);
	}
	readFile(f,word2);
	f.close();
      }
      else {
	RunSetting *rs=RunSetting::findInputSetting(word1);
	if(rs!=NULL)
	  rs->setValue(word2,lineNum,filename);
	else {
	  error << "Error: Unknown run setting name (" << word1 << ") (line " 
		<< lineNum << ") of " << filename << "." << endError;
	  quit(1);
	}
      }
    }
    else {
      error << "Error: Illegal syntax. Missing '=' (line " << lineNum << ") of " 
	    << filename << "." << endError;
      quit(1);
    }
  }
}
