//runsettings.h

#ifndef RUNSETTINGSHDR
#define RUNSETTINGSHDR

#include "util.h"

#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <vector>
#include <algorithm>

class RunSettings;

class RunSetting {
 public:
  RunSetting(string n, string d, string v, string b="", string r="") :
    name(n), doc(d), line(0), bounds(b), restriction(r), filename(""), val(v) {
    inputSettings.push_back(this);
  }
  static RunSetting *findInputSetting(const string);
  static void parseAll(const RunSettings&);
  static void printAll(ostream&);
  static void printDoc(ostream&);
  static void printDefaults(ostream&);
  void setValue(const string,int,const string);
  virtual void parse(const RunSettings&)=0;
  virtual void print(ostream&)=0;
  const string getName() const { return name; }
  const string getValue() const { return val; }

 protected:
  static vector<RunSetting*> inputSettings; // List of run settings.

  string name;           // The external name.
  string doc;		 // The documentation string.
  string val;     	 // The value.
  int line;              // The number of the line in the settings file which sets it (0=not set).
  string filename;	 // The file in which it is set.
  string bounds;         // The range.
  string restriction;    // The restriction.

  void parseError();
  void addGroup(int,vector<int>&);
  void printBadParam(const string,int,const string);
  int getGroups(string::iterator&,vector<int>&,string&,int,string&);
  void boundsError();
  void boundsParseError();
  template<class T> void checkBounds(T);
  void checkRestriction();
  int shouldPrint() const;
};

class RSInt : public RunSetting {
 public:
  int v;
  RSInt(string n, string d, string s, string b="", string r="") : RunSetting(n,d,s,b,r) {}
  void parse(const RunSettings &rs) {
    ISTRSTREAM str(val);
    if(!(str>>v))
      parseError();
    checkBounds<int>(v);
    checkRestriction();
  }
  void print(ostream& f) { f << name << "=" << v << endl; }
};

class RSDouble : public RunSetting {
 public:
  double v;
  RSDouble(string n, string d, string v, string b="", string r="") : RunSetting(n,d,v,b,r) {}
  void parse(const RunSettings &rs) {
    ISTRSTREAM str(val);
    if(!(str>>v))
      parseError();
    checkBounds<double>(v);
    checkRestriction();
  }
  void print(ostream& f) { f << name << "=" << v << endl; }
};

class RSEnum : public RunSetting {
 protected:
  const vector<string> values;
 public:
  int v;
  RSEnum(const string names[], int s, string n, string d, string v, string r="") : RunSetting(n,d,v,"",r), values(names,names+s) {}
  void parse(const RunSettings &rs) {
    ISTRSTREAM str(val);
    string s;
    if(!(str>>s))
      parseError();
    vector<string>::const_iterator p = find(values.begin(),values.end(),downcaseString(s));
      if(p==values.end())
	parseError();
    v = p - values.begin();
    checkRestriction();
  }
  void print(ostream& f) { f << name << "=" << values[v] << endl; }
};

class RSTreeType : public RSEnum {
 private:
  static const string names[];
 public:
  RSTreeType(string,string,string,string);
};

class RSBoolean : public RSEnum {
 private:
  static const string names[];
 public:
  RSBoolean(string,string,string,string);
};

class RSIntArray : public RunSetting {
 public:
  vector<int> v;
  RSIntArray(string n, string d, string v, string b="", string r="") : RunSetting(n,d,v,b,r) {}
  void parse(const RunSettings &rs) {
    ISTRSTREAM str(val);
    int n;
    while(str>>n) {
      checkBounds<int>(n);
      v.push_back(n);
      char ch;
      if(!(str>>ch) || ch!=',')
	break;
    }
    checkRestriction();
  }
  void print(ostream& f) { 
    f << name << "=";
    if(v.size()>0) {
      f << v[0];
      if(v.size()>1)
	for(int i=1;i<v.size();i++)
	  f << "," << v[i];
    }
    f << endl;
  }
};  

class RSUnsignedInt : public RunSetting {
 public:
  uint32 v;
  RSUnsignedInt(string n, string d, string s, string b="", string r="") : RunSetting(n,d,s,b,r) {}
  void parse(const RunSettings &rs) {
    ISTRSTREAM str(val);
    if(!(str>>v))
      parseError();
    checkBounds<unsigned int>(v);
    checkRestriction();
  }
  void print(ostream& f) { f << name << "=" << v << endl; }
};

class RSGroupList : public RunSetting {
 public:
  string v;                       // list of groups
  vector<int> groups;             // mapping from tag number to group number

  RSGroupList(string n, string d, string v, string r="") : RunSetting(n,d,v,"",r) {}
  void parse(const RunSettings &);
  void print(ostream& f) { f << name << "=" << v << endl; }
};

class RSString : public RunSetting {
 public:
  string v;
  RSString(string n, string d, string v, string r="") : RunSetting(n,d,v,"",r) {}
  void parse(const RunSettings &rs) { 
    checkRestriction();
    v = val; 
  }
  void print(ostream& f) { f << name << "=" << v << endl; }
};

class RSEnumList : public RunSetting {
 public:
  vector<int> v;
  RSEnumList(const string names[], int s, string n, string d, string v, string r="") :
    RunSetting(n,d,v,"",r), values(names,names+s) 
    {}
  void parse(const RunSettings &rs) {
    int p1=0,p2;
    while((p2=val.find_first_of(",",p1))!=string::npos) {
      v.push_back(getValue(val.substr(p1,p2-p1)));
      p1 = p2 + 1;
    }
    v.push_back(getValue(val.substr(p1,p2-p1)));
    checkRestriction();
  }
  void print(ostream& f) { 
    f << name << "=";
    if(v.size()>0) {
      f << values[v[0]];
      if(v.size()>1)
	for(int i=1;i<v.size();i++)
	  f << "," << values[v[i]];
    }
    f << endl;
  }
 protected:
  const vector<string> values; 
  int getValue(const string& x) {
    vector<string>::const_iterator p = find(values.begin(),values.end(),downcaseString(x));
    if(p==values.end())
      parseError();
    return p - values.begin();
  }
};  

class RSUpdateMethod : public RSEnumList {
 private:
  static const string names[];
 public:
  RSUpdateMethod(string,string,string,string);
};

class RunSettings {
 public:

  RunSettings(istream&,const string&);

  void print(ostream& c) const { RunSetting::printAll(c); }

  int getSeed() const { return seed.v; }
  int getPreBurnCycles() const { return preBurnCycles.v;}
  int getNumCycles() const { return cycles.v;}
  int getSampleInterval() const { return sampleInterval.v;}
  int getWindowInterval() const { return windowInterval.v;}
  int getTuningInterval() const { return tuningInterval.v;}
  const string getDataFile() const { return dataFile.v;}
  const string getMarkerFile() const {return markerFile.v;}
  const string getTreeFile() const { return treeFile.v;}
  const string getFileRoot() const { return fileRoot.v;}
  int getOutGroup() const { return outgroup.v;}
  int getInitialTreeType() const { return initialTreeType.v; }
  int getPrintAllTrees() const { return printAllTrees.v; }
  double getMu() const { return mu.v; }
  double getAlphaATuning() const {return alphaATuning.v; }
  double getAlphaSTuning() const {return alphaSTuning.v; }
  double getLambdaLTuning() const {return lambdaLTuning.v; }
  double getLambdaGTuning() const {return lambdaGTuning.v; }
  /*  int getDebug() const { return debug.v; }
  int getDebugCycle() const { return debugCycle.v; }
  int useMCMCMC() const { return mcmcmc.v; }
  int getNumChains() const { return numChains.v; }
  int getSwapFrequency() const { return swapFrequency.v; }
  double getTemperature() const { return temperature.v; }*/
  void doubleAlphaA() { alphaATuning.v *= 2; }  
  void doubleAlphaS() { alphaSTuning.v *= 2; }
  void doubleLambdaL() { lambdaLTuning.v *= 2; }
  void doubleLambdaG() { lambdaGTuning.v *= 2; }
  void halveAlphaA() { alphaATuning.v /= 2; }  
  void halveAlphaS() { alphaSTuning.v /= 2; }
  void halveLambdaL() { lambdaLTuning.v /= 2; }
  void halveLambdaG() { lambdaGTuning.v /= 2; }

  //  const vector<int> &getUpdateList() const { return updateList.v; }

  // This kludge is necessary to be consistent with the old code.
  // The order should agree with the arrays in the corresponding classes.

  enum { RANDOM, GOOD, FILE };

 private:
  RSUnsignedInt  seed;                          // odd unsigned int between 0 and (4294967295 = 2^32 - 1)
  RSInt          preBurnCycles;                 // number of preBurn cycles
  RSInt          cycles;                        // number of cycles
  RSInt          sampleInterval;                // interval for storing the tree
  RSInt          windowInterval;                // interval for printing to the terminal  
  RSInt          tuningInterval;                // interval for changing the tuning parameter
  RSString       dataFile;                      // name of file in which data is stored
  RSString       markerFile;                    // name of file in which marker lengths are stored
  RSInt          outgroup;                      // outgroup used to root the tree for printing
  RSBoolean      printAllTrees;                 // whether or not to print all the trees
  RSString       fileRoot;                      // prefix for names of output files
  RSTreeType     initialTreeType;               // source of the input tree
  RSString       treeFile;                      // file containing initial tree
  RSDouble	 mu;				// expected edge length
  RSDouble	 alphaATuning;			// tuning parameter for updating all edge lengths
  RSDouble	 alphaSTuning;			// tuning parameter for updating single edge length
  RSDouble	 lambdaLTuning;			// tuning parameter for stretching factor in local deformation algorithm
  RSDouble	 lambdaGTuning;			// tuning parameter for stretching factor in generalized local deformation algorithm
  //  RSBoolean	 debug;				// whether or not to print debugging statements
  //  RSInt		 debugCycle;			// cycle at which to being printing debugging statements
  //  RSUpdateMethod updateList;			// list of update methods to use
  //  RSBoolean      mcmcmc;                        // whether or not to use MCMCMC
  //  RSInt          numChains;                     // number of chains for MCMC
  //  RSInt		 swapFrequency;			// frequency for swapping chains in MCMCMC
  //  RSDouble	 temperature;			// temperature for MCMCMC

  void readFile(istream&,const string);
};

#endif
